"""
Routines for testing or asserting that a value is certain type, in a way that
works in both Python 2.7 and Python 3.4.
"""
from __future__ import print_function, division, absolute_import

# In Python 3, the "unicode" and "long" types went away.
string_types = (str,)  # Python 2 and 3.
try: string_types += (unicode,) # Python 2.
except NameError: pass  # Python 3.

integer_types = (int,)  # Python 2 and 3.
try: integer_types += (long,)  # Python 2.
except NameError: pass  # Python 3.

def is_string(obj):
    """
    Is the object a string?
    """
    return isinstance(obj, string_types)
    
def is_integer(obj):
    """
    Is the object an integer?
    """
    return isinstance(obj, integer_types)

def require_string(obj, name=None, nonempty=False):
    """
    Raise an exception if the obj is not of type str or unicode.
    
    If name is provided it is used in the exception message.

    If nonempty=True, then an exception is raised if the object is the empty
    string.
    """
    if not is_string(obj):
        raise TypeError(
            (("%s: " % name) if name else "") +
            "expected string, got: '%s' of type '%s'"
                % (str(obj), type(obj)))
    if nonempty and not obj:
        raise ValueError(
            (("%s: " % name) if name else "") +
            "string must be nonempty.")

def require_integer(obj, name=None):
    """
    Raise an exception if the obj is not of type int or long.
    
    If name is provided it is used in the exception message.
    """
    if not is_integer(obj):
        raise TypeError(
            (("%s: " % name) if name else "") +
            "expected int, got: '%s' of type '%s'"
                % (str(obj), type(obj)))





from __future__ import print_function, division, absolute_import

from .compute_cache import clear_cached_objects
from .ensembl_release import EnsemblRelease
from .gene import Gene
from .gtf import GTF
from .locus import Locus
from .reference_transcripts import ReferenceTranscripts
from .search import find_nearest_locus
from .transcript import Transcript



"""
Helper functions for searching over collections of PyEnsembl objects
"""

def find_nearest_locus(start, end, loci):
    """
    Finds nearest locus (object with method `distance_to_interval`) to the
    interval defined by the given `start` and `end` positions.
    Returns the distance to that locus, along with the locus object itself.
    """
    best_distance = float("inf")
    best_locus = None
    for locus in loci:
        distance = locus.distance_to_interval(start, end)

        if best_distance > distance:
            best_distance = distance
            best_locus = locus

    return best_distance, best_locus



from __future__ import print_function, division, absolute_import

from os.path import join

from .common import CACHE_SUBDIR
from .release_info import which_human_reference_name, check_release_number
from .url_templates import ENSEMBL_FTP_SERVER, fasta_cdna_url_parts

import pyfaidx
from datacache import Cache

class ReferenceTranscripts(object):
    """
    Container for reference genome sequenes. Downloads and caches reference
    FASTA files.
    Currently only supports (unspliced) transcript sequences.

    TODO: Access introns and intergenic regions by also downloading
    the *.dna.fa.gz FASTA files. Should rename class to Reference when
    this gets implemented.
    """
    def __init__(
            self,
            ensembl_release,
            species="homo_sapiens",
            server=ENSEMBL_FTP_SERVER):

        # download cache for fetching reference FASTA files
        self.cache = Cache(CACHE_SUBDIR)

        self.release = check_release_number(ensembl_release)

        assert species == "homo_sapiens", \
            "Species other than human not supported: %s" % (species,)
        self.reference_name = which_human_reference_name(ensembl_release)
        assert species == "homo_sapiens", \
            "Species %s not currently supported" % (species,)
        self.species = species

        self.server = server

        reference_url_dir, reference_filename = fasta_cdna_url_parts(
            ensembl_release=self.release,
            species=self.species,
            server=server)
        self.remote_filename = reference_filename
        self.url = join(reference_url_dir, reference_filename)

        # dictionary mapping transcript IDs to cDNA sequences
        self._transcript_sequences = {}

        # pyfaidx Fasta dictionary gets lazily constructed
        # since we only want to download the Fasta file if/when
        # we use it
        self._fasta_dictionary = None

        # key set for fasta dictionary, for faster membership tests,
        # pyfaidx does something upsettingly slow in its __contains__
        # method, see comment on `transcript_sequence`.
        self._fasta_keys = None

    def __str__(self):
        return "ReferenceTranscripts(release=%s, species=%s, filename=%s)"  % (
            self.release, self.species, self.remote_filename)

    def __repr__(self):
        return str(self)

    def __contains__(self, transcript_id):
        # the pyfaidx __contains__ method requires an expensive list traversal
        # so cache the keys as a set for faster membership checks
        if self._fasta_keys is None:
            keys = list(self.fasta_dictionary.keys())
            self._fasta_keys = set(keys)
        return transcript_id in self._fasta_keys

    def __eq__(self, other):
        return (
            isinstance(other, ReferenceTranscripts) and
            self.ensembl_release == other.ensembl_release and
            self.species == other.species and
            self.server == other.server)

    @property
    def local_fasta_path(self):
        """
        Returns local path to FASTA file,
        download from the Ensembl FTP server if not already cached.
        """
        return self.cache.fetch(self.url, self.remote_filename, decompress=True)

    @property
    def local_dir(self):
        return split(self.local_fasta_path)[0]

    @property
    def local_filename(self):
        return split(self.local_fasta_path)[1]

    @property
    def fasta_dictionary(self):
        if not self._fasta_dictionary:
            self._fasta_dictionary = pyfaidx.Fasta(self.local_fasta_path)
        return self._fasta_dictionary

    def transcript_sequence(self, transcript_id):
        if transcript_id not in self._transcript_sequences:
            if not transcript_id.startswith("ENST"):
                raise ValueError("Invalid transcript ID: %s" % (transcript_id,))

            # the __getitem__ on pyfaidx.Fasta does a list copy and
            # traversal to check whether the transcript ID is valid,
            # much faster if by-pass this logic with our own key membershop
            # check and then construct the FastaRecord ourselves.
            #
            # Example speedup: previously 24s, now 3s annotating 311 transcripts
            #
            if transcript_id not in self:
                raise ValueError(
                    "Transcript ID not found: %s" % (transcript_id,))

            fasta_record = pyfaidx.FastaRecord(
                transcript_id, self.fasta_dictionary)

            # FastaRecord doesn't seem to have an accessor to get the full
            # sequence (only subsequences), so slice out the full string
            seq = fasta_record[:len(fasta_record)]
            self._transcript_sequences[transcript_id] = seq
        return self._transcript_sequences[transcript_id]




from test_common import test_ensembl_releases
from pyensembl import EnsemblRelease, find_nearest_locus
from nose.tools import eq_

@test_ensembl_releases
def test_find_nearest_BRAF_exon(ensembl):
    braf = ensembl.genes_by_name("BRAF")[0]
    braf_transcripts = braf.transcripts
    for exon in braf_transcripts[0].exons:
        # immediately before exon
        result = find_nearest_locus(
            start=exon.start-2,
            end=exon.end-1,
            loci=exons)
        eq_(result, (1, exon))

        # overlapping with exon
        result = find_nearest_locus(
            start=exon.start-2,
            end=exon.start+1,
            loci=exons)
        eq_(result, (0, exon))

        # immediately after exon
        result = find_nearest_locus(
            start=exon.end+1,
            end=exon.end+2,
            loci=exons)
        eq_(result, (1, exon))

@test_ensembl_releases
def test_find_nearest_BRAF_transcript(ensembl):
    braf = ensembl.genes_by_name("BRAF")[0]
    braf_transcripts = braf.transcripts
    for transcript in braf_transcripts:
        # immediately before transcript
        result = find_nearest_locus(
            start=transcript.start-2,
            end=transcript.end-1,
            loci=braf_transcripts)
        eq_(result, (1, transcript))

        # overlapping with transcript
        result = find_nearest_locus(
            start=transcript.start-2,
            end=transcript.start+1,
            exons=braf_transcripts)
        eq_(result, (0, transcript))

        # immediately after transcript
        result = find_nearest_locus(
            start=transcript.end+1,
            end=transcript.end+2,
            exons=braf_transcripts)
        eq_(result, (1, transcript))


from __future__ import print_function, division, absolute_import

from .biotypes import is_valid_biotype
from .locus import Locus
from .transcript import Transcript
from .type_checks import require_integer, require_string

from memoized_property import memoized_property

class Gene(Locus):

    def __init__(self, gene_id, db, reference):
        require_string(gene_id, "gene ID")

        self.id = gene_id
        self.db = db
        self.reference = reference
        columns = [
            'gene_name',
            'seqname',
            'start',
            'end',
            'strand',
            'gene_biotype'
        ]
        gene_name, contig, start, end, strand, biotype = self.db.query_one(
            columns,
            filter_column='gene_id',
            filter_value=gene_id,
            feature='gene')
        if not gene_name:
            raise ValueError("Missing name for gene with ID = %s" % gene_id)
        self.name = gene_name

        Locus.__init__(self, contig, start, end, strand)

        if not biotype:
            raise ValueError(
                "Missing gene_biotype for gene with ID = %s" % gene_id)
        elif not is_valid_biotype(biotype):
            raise ValueError(
                "Invalid gene_biotype %s for gene with ID = %s" % (
                    biotype, gene_id))
        self.biotype = biotype

    def __str__(self):
        return "Gene(id=%s, name=%s)" % (self.id, self.name)

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return (
            isinstance(other, Gene) and
            self.id == other.id and
            self.db == other.db and
            self.reference == other.reference)

    def __hash__(self):
        return hash(self.id)

    @memoized_property
    def transcripts(self):
        """
        Property which dynamically construct transcript objects for all
        transcript IDs associated with this gene.
        """
        transcript_id_results = self.db.query(
            select_column_names=['transcript_id'],
            filter_column='gene_id',
            filter_value=self.id,
            feature='transcript',
            distinct=False,
            required=False)

        # We're doing a SQL query for each transcript ID to fetch
        # its particular information, might be more efficient if we
        # just get all the columns here, but how do we keep that modular?
        return [
            Transcript(result[0], self.db, self.reference)
            for result in transcript_id_results
        ]

    @memoized_property
    def exons(self):
        exons_dict = {}
        for transcript in self.transcripts:
            for exon in transcript.exons:
                if exon.id not in exons_dict:
                    exons_dict[exon.id] = exon
        return list(exons_dict.values())

